==> cpl_6.0.orig.tar.gz: cpl-6.0/cpldrs/cpl_fft.c <==
/* $Id: cpl_fft.c,v 1.21 2012/03/12 09:48:27 llundin Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2008 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


/*----------------------------------------------------------------------------
                                   Includes
 ----------------------------------------------------------------------------*/

/* Must be included first to ensure declaration of complex image accessors */
#include <complex.h>

#include "cpl_fft.h"

#include "cpl_error_impl.h"

#if defined CPL_FFTWF_INSTALLED || defined CPL_FFTW_INSTALLED
/* If FFTW is installed */
#include <fftw3.h>
#endif

#include <math.h>
#include <string.h>

/*---------------------------------------------------------------------------*/
/**
 * @defgroup cpl_fft FFTW wrappers
 *
 * This module provides FFTW wrappers
 *
 * @par Synopsis:
 * @code
 *   #include "cpl_fft.h"
 * @endcode
 */
/*---------------------------------------------------------------------------*/
/**@{*/

/*-----------------------------------------------------------------------------
                                   Private functions
 -----------------------------------------------------------------------------*/

#ifdef CPL_FFTW_INSTALLED
#  define CPL_FFTW fftw
/* Cannot concatenate the reserved macro complex :-( */
#  define CPL_FFTW_TYPE fftw_complex
#  define CPL_TYPE double
/* Cannot concatenate the reserved macro complex :-( */
#  define CPL_TYPE_C double_complex
#  include "cpl_fft_body.h"
#  undef CPL_FFTW
#  undef CPL_TYPE
#  undef CPL_TYPE_T
#  undef CPL_TYPE_C
#  undef CPL_FFTW_TYPE
#endif

#ifdef CPL_FFTWF_INSTALLED
#  define CPL_FFTW fftwf
/* Cannot concatenate the reserved macro complex :-( */
#  define CPL_FFTW_TYPE fftwf_complex
#  define CPL_TYPE float
/* Cannot concatenate the reserved macro complex :-( */
#  define CPL_TYPE_C float_complex
#  include "cpl_fft_body.h"
#  undef CPL_FFTW
#  undef CPL_TYPE
#  undef CPL_TYPE_T
#  undef CPL_TYPE_C
#  undef CPL_FFTW_TYPE
#endif

/*---------------------------------------------------------------------------*/
/**
  @brief    Perform a FFT operation on an image
  @param  self  Pre-allocated output image to transform to
  @param  other Input image to transform from
  @param  mode  CPL_FFT_FORWARD or CPL_FFT_BACKWARD, optionally CPL_FFT_NOSCALE
  @return CPL_ERROR_NONE or the corresponding #_cpl_error_code_ on error
  
  This function performs an FFT on an image, using FFTW. CPL may be configured
  without this library, in this case an otherwise valid call will set and return
  the error CPL_ERROR_UNSUPPORTED_MODE.

  The input and output images must match in precision level. Integer images are
  not supported.

  In a forward transform the input image may be non-complex. In this case a
  real-to-complex transform is performed. This will only compute the first
  ny/2 + 1 columns of the transform. In this transform it is allowed to pass
  an output image with ny/2 + 1 columns.

  Similarly, in a backward transform the output image may be non-complex. In
  this case a complex-to-real transform is performed. This will only transform
  the first ny/2 + 1 columns of the input. In this transform it is allowed to
  pass an input image with ny/2 + 1 columns.

  Per default the backward transform scales (divides) the result with the
  number of elements transformed (i.e. the number of pixels in the result
  image). This scaling can be turned off with CPL_FFT_NOSCALE.

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if an image is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the mode is illegal
  - CPL_ERROR_INCOMPATIBLE_INPUT if the image sizes do not match
  - CPL_ERROR_TYPE_MISMATCH if the image types are incompatible with each other
                            or with the transform
  - CPL_ERROR_UNSUPPORTED_MODE if FFTW has not been installed
 */
/*---------------------------------------------------------------------------*/
cpl_error_code cpl_fft_image(cpl_image * self, const cpl_image * other,
                             cpl_fft_mode mode)
{
    const cpl_type typin  = cpl_image_get_type(other);
    const cpl_type typout = cpl_image_get_type(self);

    const cpl_size lnxin  = cpl_image_get_size_x(other);
    const cpl_size lnyin  = cpl_image_get_size_y(other);
    const cpl_size lnxout = cpl_image_get_size_x(self);
    const cpl_size lnyout = cpl_image_get_size_y(self);

    const int      nxin   = (int)lnxin;
    const int      nyin   = (int)lnyin;
    const int      nxout  = (int)lnxout;
    const int      nyout  = (int)lnyout;
    const int      nxh    = ((mode & CPL_FFT_FORWARD) ? nxin : nxout) / 2 + 1;

    cpl_error_code error;


    cpl_ensure_code(self  != NULL,             CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(other    != NULL,             CPL_ERROR_NULL_INPUT);

    /* FFTW (only) supports dimensions of type int */
    cpl_ensure_code((cpl_size)nxin  == lnxin,  CPL_ERROR_UNSUPPORTED_MODE);
    cpl_ensure_code((cpl_size)nyin  == lnyin,  CPL_ERROR_UNSUPPORTED_MODE);
    cpl_ensure_code((cpl_size)nxout == lnxout, CPL_ERROR_UNSUPPORTED_MODE);
    cpl_ensure_code((cpl_size)nyout == lnyout, CPL_ERROR_UNSUPPORTED_MODE);

    if (mode & CPL_FFT_FORWARD && !(typin & CPL_TYPE_COMPLEX)) {
        cpl_ensure_code(nxin == nxout || nxout == nxh,
                        CPL_ERROR_INCOMPATIBLE_INPUT);
    } else if (mode & CPL_FFT_BACKWARD && !(typout & CPL_TYPE_COMPLEX)) {
        cpl_ensure_code(nxin == nxout || nxin == nxh,
                        CPL_ERROR_INCOMPATIBLE_INPUT);
    } else {
        cpl_ensure_code(nxin == nxout,         CPL_ERROR_INCOMPATIBLE_INPUT);
    }
    cpl_ensure_code(lnyin == lnyout,           CPL_ERROR_INCOMPATIBLE_INPUT);

   if (typin & CPL_TYPE_FLOAT) {

       cpl_ensure_code(typout & CPL_TYPE_FLOAT, CPL_ERROR_TYPE_MISMATCH);

#ifdef CPL_FFTWF_INSTALLED
       error = cpl_fft_image_float(self, other, mode);
#else
       error = CPL_ERROR_UNSUPPORTED_MODE;
#endif
   } else if (typin & CPL_TYPE_DOUBLE) {

       cpl_ensure_code(typout & CPL_TYPE_DOUBLE, CPL_ERROR_TYPE_MISMATCH);

#ifdef CPL_FFTW_INSTALLED
       error = cpl_fft_image_double(self, other, mode);
#else
       error = CPL_ERROR_UNSUPPORTED_MODE;
#endif
   } else {
       error = CPL_ERROR_TYPE_MISMATCH;
   }

   /* Set or propagate error, if any */
   return cpl_error_set_message_(error, "mode=%d. %d X %d (%s) => "
                                 "%d X %d (%s)", (int)mode,
                                 nxin, nyin, cpl_type_get_name(typin),
                                 nxout, nyout, cpl_type_get_name(typout));
}

/*----------------------------------------------------------------------------*/
/**
  @brief  Perform a FFT operation on the images in an imagelist
  @param  self  Pre-allocated output imagelist to transform to
  @param  other Input imagelist to transform from
  @param  mode  CPL_FFT_FORWARD or CPL_FFT_BACKWARD, optionally CPL_FFT_NOSCALE
  @return CPL_ERROR_NONE or the corresponding #_cpl_error_code_ on error
  @see cpl_fft_image()

 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_fft_imagelist(cpl_imagelist       * self,
                                 const cpl_imagelist * other,
                                 cpl_fft_mode          mode)
{
    const cpl_size sizein = cpl_imagelist_get_size(other);
    cpl_size i;

    cpl_ensure_code(self != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(other  != NULL, CPL_ERROR_NULL_INPUT);
        
    cpl_ensure_code(cpl_imagelist_get_size(self) == sizein,
                    CPL_ERROR_INCOMPATIBLE_INPUT);

    for (i = 0; i < sizein; i++) {
        cpl_image       * imgout = cpl_imagelist_get(self, i);
        const cpl_image * imgin  = cpl_imagelist_get_const(other, i);

        if (cpl_fft_image(imgout, imgin, mode)) break;
    }

    return i == sizein ? CPL_ERROR_NONE : cpl_error_set_where_();
}

/**@}*/

==> cpl_6.4.orig.tar.gz: cpl-6.4/cpldrs/cpl_fft.c <==
/* $Id: cpl_fft.c,v 1.34 2012-05-11 15:44:26 llundin Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2008 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


/*----------------------------------------------------------------------------
                                   Includes
 ----------------------------------------------------------------------------*/

/* Must be included first to ensure declaration of complex image accessors */
#include <complex.h>

#include "cpl_fft.h"

#include "cpl_error_impl.h"

#if defined CPL_FFTWF_INSTALLED || defined CPL_FFTW_INSTALLED
/* If FFTW is installed */
#include <fftw3.h>
#endif

#include <string.h>

/*---------------------------------------------------------------------------*/
/**
 * @defgroup cpl_fft FFTW wrappers
 *
 * This module provides FFTW wrappers
 *
 * @par Synopsis:
 * @code
 *   #include "cpl_fft.h"
 * @endcode
 */
/*---------------------------------------------------------------------------*/
/**@{*/

/*-----------------------------------------------------------------------------
                                   Private functions
 -----------------------------------------------------------------------------*/

#if defined CPL_FFTWF_INSTALLED || defined CPL_FFTW_INSTALLED
static void * cpl_fft_aligned(void *, void *, size_t) CPL_ATTR_NONNULL;
#endif

static cpl_error_code cpl_fft_image_(cpl_image *, const cpl_image *,
                                     cpl_fft_mode, void *, void *, void *,
                                     cpl_boolean)

#ifdef CPL_HAVE_ATTR_NONNULL
    __attribute__((nonnull(5,6)))
#endif
    ;

#ifdef CPL_FFTW_INSTALLED
#  define CPL_FFTW fftw
/* Cannot concatenate the reserved macro complex :-( */
#  define CPL_FFTW_TYPE fftw_complex
#  define CPL_TYPE double
/* Cannot concatenate the reserved macro complex :-( */
#  define CPL_TYPE_C double_complex
#  include "cpl_fft_body.h"
#  undef CPL_FFTW
#  undef CPL_TYPE
#  undef CPL_TYPE_T
#  undef CPL_TYPE_C
#  undef CPL_FFTW_TYPE
#endif

#ifdef CPL_FFTWF_INSTALLED
#  define CPL_FFTW fftwf
/* Cannot concatenate the reserved macro complex :-( */
#  define CPL_FFTW_TYPE fftwf_complex
#  define CPL_TYPE float
/* Cannot concatenate the reserved macro complex :-( */
#  define CPL_TYPE_C float_complex
#  include "cpl_fft_body.h"
#  undef CPL_FFTW
#  undef CPL_TYPE
#  undef CPL_TYPE_T
#  undef CPL_TYPE_C
#  undef CPL_FFTW_TYPE
#endif

/*---------------------------------------------------------------------------*/
/**
  @brief    Perform a FFT operation on an image
  @param  self  Pre-allocated output image to transform to
  @param  other Input image to transform from, use self for in-place transform
  @param  mode  CPL_FFT_FORWARD or CPL_FFT_BACKWARD, optionally CPL_FFT_NOSCALE
  @return CPL_ERROR_NONE or the corresponding #_cpl_error_code_ on error
  
  This function performs an FFT on an image, using FFTW. CPL may be configured
  without this library, in this case an otherwise valid call will set and return
  the error CPL_ERROR_UNSUPPORTED_MODE.

  The input and output images must match in precision level. Integer images are
  not supported.

  In a forward transform the input image may be non-complex. In this case a
  real-to-complex transform is performed. This will only compute the first
  ny/2 + 1 columns of the transform. In this transform it is allowed to pass
  an output image with ny/2 + 1 columns.

  Similarly, in a backward transform the output image may be non-complex. In
  this case a complex-to-real transform is performed. This will only transform
  the first ny/2 + 1 columns of the input. In this transform it is allowed to
  pass an input image with ny/2 + 1 columns.

  Per default the backward transform scales (divides) the result with the
  number of elements transformed (i.e. the number of pixels in the result
  image). This scaling can be turned off with CPL_FFT_NOSCALE.

  If many transformations in the same direction are to be done on data of the
  same size and type, a reduction in the time required to perform the
  transformations can be achieved by adding the flag CPL_FFT_FIND_MEASURE
  to the first transformation.
  For a larger number of transformations a further reduction may be achived
  with the flag CPL_FFT_FIND_PATIENT and for an even larger number of
  transformations a further reduction may be achived with the flag
  CPL_FFT_FIND_EXHAUSTIVE.

  If many transformations are to be done then a reduction in the time required
  to perform the transformations can be achieved by using cpl_fft_imagelist().  

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if an image is NULL
  - CPL_ERROR_ILLEGAL_INPUT if the mode is illegal
  - CPL_ERROR_INCOMPATIBLE_INPUT if the image sizes do not match
  - CPL_ERROR_TYPE_MISMATCH if the image types are incompatible with each other
                            or with the transform
  - CPL_ERROR_UNSUPPORTED_MODE if FFTW has not been installed
 */
/*---------------------------------------------------------------------------*/
cpl_error_code cpl_fft_image(cpl_image * self, const cpl_image * other,
                             cpl_fft_mode mode)
{

    /*
      The below pointers hold temporary storage allocated and deallocated by
      cpl_fft_image_().
    */
    void * bufin  = NULL;
    void * bufout = NULL;

    return cpl_fft_image_(self, other, mode, NULL, &bufin, &bufout, CPL_TRUE)
        ? cpl_error_set_where_() : CPL_ERROR_NONE;

}

/*---------------------------------------------------------------------------*/
/**
  @internal
  @brief Perform a FFT operation on an image
  @param self    Pre-allocated output image to transform to
  @param other   Input image to transform from, use self for in-place transform
  @param mode    CPL_FFT_FORWARD or CPL_FFT_BACKWARD, optionally CPL_FFT_NOSCALE
  @param pplan   NULL, or a pointer to keep the plan
  @param pbufin  A pointer to keep the input buffer
  @param pbufout A pointer to keep the output buffer
  @param is_last CPL_TRUE for the last call with the given pplan
  @return CPL_ERROR_NONE or the corresponding #_cpl_error_code_ on error
  @see cpl_fft_image()
 */
/*---------------------------------------------------------------------------*/
static cpl_error_code cpl_fft_image_(cpl_image * self, const cpl_image * other,
                                     cpl_fft_mode mode, void * pplan,
                                     void * pbufin, void * pbufout,
                                     cpl_boolean is_last)
{
    const cpl_type typin  = cpl_image_get_type(other);
    const cpl_type typout = cpl_image_get_type(self);

    const cpl_size lnxin  = cpl_image_get_size_x(other);
    const cpl_size lnyin  = cpl_image_get_size_y(other);
    const cpl_size lnxout = cpl_image_get_size_x(self);
    const cpl_size lnyout = cpl_image_get_size_y(self);

    const int      nxin   = (int)lnxin;
    const int      nyin   = (int)lnyin;
    const int      nxout  = (int)lnxout;
    const int      nyout  = (int)lnyout;
    const int      nxh    = ((mode & CPL_FFT_FORWARD) ? nxin : nxout) / 2 + 1;

    cpl_error_code error;
#if defined CPL_FFTWF_INSTALLED || defined CPL_FFTW_INSTALLED
    unsigned rigor = FFTW_ESTIMATE; /* Default value */


    if (mode & CPL_FFT_FIND_EXHAUSTIVE) {
        rigor = FFTW_EXHAUSTIVE;
        /* Reset bit. At the end the bitmask must be zero */
        mode ^= CPL_FFT_FIND_EXHAUSTIVE;
    } else if (mode & CPL_FFT_FIND_PATIENT) {
        rigor = FFTW_PATIENT;
        /* Reset bit. At the end the bitmask must be zero */
        mode ^= CPL_FFT_FIND_PATIENT;
    } else if (mode & CPL_FFT_FIND_MEASURE) {
        rigor = FFTW_MEASURE;
        /* Reset bit. At the end the bitmask must be zero */
        mode ^= CPL_FFT_FIND_MEASURE;
    }

#endif

    cpl_ensure_code(self  != NULL,             CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(other != NULL,             CPL_ERROR_NULL_INPUT);

    /* FFTW (only) supports dimensions that fit within an int type */
    cpl_ensure_code((cpl_size)nxin  == lnxin,  CPL_ERROR_UNSUPPORTED_MODE);
    cpl_ensure_code((cpl_size)nyin  == lnyin,  CPL_ERROR_UNSUPPORTED_MODE);
    cpl_ensure_code((cpl_size)nxout == lnxout, CPL_ERROR_UNSUPPORTED_MODE);
    cpl_ensure_code((cpl_size)nyout == lnyout, CPL_ERROR_UNSUPPORTED_MODE);

    if (mode & CPL_FFT_FORWARD && !(typin & CPL_TYPE_COMPLEX)) {
        cpl_ensure_code(nxin == nxout || nxout == nxh,
                        CPL_ERROR_INCOMPATIBLE_INPUT);
    } else if (mode & CPL_FFT_BACKWARD && !(typout & CPL_TYPE_COMPLEX)) {
        cpl_ensure_code(nxin == nxout || nxin == nxh,
                        CPL_ERROR_INCOMPATIBLE_INPUT);
    } else {
        cpl_ensure_code(nxin == nxout,         CPL_ERROR_INCOMPATIBLE_INPUT);
    }
    cpl_ensure_code(lnyin == lnyout,           CPL_ERROR_INCOMPATIBLE_INPUT);

   if (typin & CPL_TYPE_FLOAT) {

       cpl_ensure_code(typout & CPL_TYPE_FLOAT, CPL_ERROR_TYPE_MISMATCH);

#ifdef CPL_FFTWF_INSTALLED
       error = cpl_fft_image_float(self, other, mode, rigor, (fftwf_plan*)pplan,
                                   (fftwf_complex**)pbufin,
                                   (fftwf_complex**)pbufout, is_last);
#else
       error = CPL_ERROR_UNSUPPORTED_MODE;
#endif
   } else if (typin & CPL_TYPE_DOUBLE) {

       cpl_ensure_code(typout & CPL_TYPE_DOUBLE, CPL_ERROR_TYPE_MISMATCH);

#ifdef CPL_FFTW_INSTALLED
       error = cpl_fft_image_double(self, other, mode, rigor, (fftw_plan*)pplan,
                                    (fftw_complex**)pbufin,
                                    (fftw_complex**)pbufout, is_last);
#else
       error = CPL_ERROR_UNSUPPORTED_MODE;
#endif
   } else {
       error = CPL_ERROR_TYPE_MISMATCH;
   }

   /* Set or propagate error, if any */
   return cpl_error_set_message_(error, "mode=%d (%d). %d X %d (%s) => %d X"
                                 " %d (%s)", (int)mode, (int)is_last,
                                 nxin, nyin, cpl_type_get_name(typin),
                                 nxout, nyout, cpl_type_get_name(typout));
}

/*----------------------------------------------------------------------------*/
/**
  @brief  Perform a FFT operation on the images in an imagelist
  @param  self  Pre-allocated output imagelist to transform to
  @param  other Input imagelist to transform from
  @param  mode  CPL_FFT_FORWARD or CPL_FFT_BACKWARD, optionally CPL_FFT_NOSCALE
  @return CPL_ERROR_NONE or the corresponding #_cpl_error_code_ on error
  @see cpl_fft_image()

 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_fft_imagelist(cpl_imagelist       * self,
                                 const cpl_imagelist * other,
                                 cpl_fft_mode          mode)
{
    const cpl_size    sizein   = cpl_imagelist_get_size(other);
    /*
      The below pointers hold temporary storage allocated by
      cpl_fft_image_(), the first call (if successful) allocates and
      the last call is responsible for the deallocation (so if the first
      call succeeds the subsequent call(s) are assumed to succeed).

    */
    void * plan   = NULL;
    void * bufin  = NULL;
    void * bufout = NULL;
    cpl_size i;

    cpl_ensure_code(self  != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(other != NULL, CPL_ERROR_NULL_INPUT);
        
    cpl_ensure_code(cpl_imagelist_get_size(self) == sizein,
                    CPL_ERROR_INCOMPATIBLE_INPUT);

    for (i = 0; i < sizein; i++) {
        cpl_image       * imgout = cpl_imagelist_get(self, i);
        const cpl_image * imgin  = cpl_imagelist_get_const(other, i);

        if (cpl_fft_image_(imgout, imgin, mode, &plan,
                           &bufin, &bufout, i+1 == sizein)) break;
    }

    return i == sizein ? CPL_ERROR_NONE : cpl_error_set_where_();
}

/**@}*/

#if defined CPL_FFTWF_INSTALLED || defined CPL_FFTW_INSTALLED
/*----------------------------------------------------------------------------*/
/**
  @brief  Return CPL_TRUE on proper alignment (4 bits or matching align mask)
  @param  self  Pointer to return, if aligned
  @param  other Aligned pointer, returned if self is not aligned
  @param  align Pattern
  @return Aligned pointer

 */
/*----------------------------------------------------------------------------*/
static void * cpl_fft_aligned(void * self, void * other, size_t align)
{
    const size_t myself = (const size_t)self;
    size_t lsb = 1;

    /* Find number of zero-valued LSBs */
    while (!(align & lsb) && lsb < 16) {
        lsb <<=1;
        lsb |= 1;
    }
    lsb >>= 1;

    /* Iff self has a 1 where lsb has a 1, then self is not aligned */
    return (myself & lsb) ? other : self;
}
#endif
